import json
import logging
import os
import shutil
import sys
import urllib2

from . import ClientException
from .file_downloader import FileDownloader
from .patcher import Patcher
from .utils import ChDir, rsa_verify

log = logging.getLogger(__name__)


class Client(object):
    """Used on client side to update files

    Kwargs:

        obj (instance): config object
    """

    def __init__(self, obj=None):
        if obj:
            self.init_app(obj)

    def init_app(self, obj):
        """Sets up client with config values from obj

        Args:
            obj (instance): config object

        """
        self.update_url = self._fix_update_url(obj.config.get('UPDATE_URL'))
        self.app_name = obj.config.get('APP_NAME')
        self.data_dir = obj.config.get('APP_DATA_DIR', None)
        self.public_key = obj.config.get('PUBLIC_KEY')
        self.debug = obj.config.get('DEBUG', False)
        self.version_file = 'version.json'
        self.version_url = self.update_url + self.version_file
        self.json_data = None
        self.verified = False
        if self.data_dir is None:
            extra = '.'
            if sys.platform == 'win32':
                extra = ''
            self.data_dir = os.path.join(os.path.expanduser('~'),
                                         extra + self.app_name)
        self.update_folder = os.path.join(self.data_dir, 'update')
        self._setup()

    def update_check(self, name, version):
        """Proxy method for :meth:`_patch_update` & :meth:`_full_update`. Will try to
        patch binary if all check pass.  IE hash verified signature
        verified.  If any check doesn't pass then falls back to full
        update

        Args:
            name (str): Name of file to update

            version (str): Current version number of file to update

        Returns:
            (bool) Meanings::

                True - Update Successful

                False - Update Failed
        """
        if not self.verified:
            log.error('Version file not verified')
            return
        log.info('Checking for {} updates...'.format(name))
        patch_success = self._patch_update(name, version)
        if patch_success:
            log.info('Successful patch update')
            return

        update_success = self._full_update(name, version)
        if update_success:
            return True
        else:
            return False

    def _patch_update(self, name, version):
        """Updates the binary by patching

        Args:
            name (str): Name of file to update

            version (str): Current version number of file to update

        Returns:
            (bool) Meanings::

                True - Either already up to date or patched successful

                False - Either failed to patch or no base binary to patch

        """
        filename = self.json_data['y_data'][name][version]['filename']
        highest_version = self._get_highest_version(name)
        if highest_version <= version:
            log.info('{} already updated to the latest version'.format(name))
            return True

        # Just checking to see if the latest update is available to patch
        # If not we'll just do a full binary download
        if not os.path.exists(os.path.join(self.update_folder, filename)):
            log.warning('{} got deleted. No base to start patching form'.format(filename))
            return False

        p = Patcher(name, self.json_data, version, highest_version,
                    self.update_folder)

        # Returns True if everything went well
        result = p.start()
        log.info('Was patch update successful? {}'.format(result))
        return result


    def _full_update(self, name, version):
        """Updates the binary by downloading full update

        Args:
            name (str): Name of file to update

            version (str): Current version number of file to update

        Returns:
            (bool) Meanings::

                True - Update Successful

                False - Update Failed
        """
        # Takes resource name and current installed version number
        #
        # Does the actual heavy lifting.  Gets highest version
        # number then compares with pass version arg.
        highest_version = self._get_highest_version(name)
        if highest_version <= version:
            log.info('{} already updated to the latest version'.format(name))
            return

        url = self._get_file_url(name, highest_version)
        filename = self._get_filename_with_extension(name,
                                                     highest_version)
        log.info('Starting full update')
        with ChDir(self.update_folder):
            # _hash = self.json_data['y_data'][name][version]['md5']
            _hash = self.json_data['y_data'][name][highest_version]['md5']
            log.info('Downloading update')
            fd = FileDownloader(filename, url, _hash)
            result = fd.download_verify_write()
            if result:
                log.info('Update Complete')
                return True
            else:
                log.error('Failed To Updated To Latest Version')
                return False

    def _setup(self):
        """Sets up required directories on end-users computer
        to place verified update data
        """
        log.debug('Setting up directories...')
        dirs = [self.data_dir, self.update_folder]
        for d in dirs:
            if not os.path.exists(d):
                log.debug('Creating directory: {}'.format(d))
                os.mkdir(d)
        self._load_json()
        self._check_version_file_sig()

    def _check_version_file_sig(self):
        """Verifies version file signature"""
        if 'sig' in self.json_data.keys():
            self.sig = self.json_data['sig']
            log.info('Deleting sig from update data')
            del self.json_data['sig']
            update_data = json.dumps(self.json_data, sort_keys=True)
            result = rsa_verify(update_data, self.sig, self.public_key)
            if result:
                log.info('Version file verified')
                self.verified = True
            else:
                log.warning('Version file sig mismatch')
        else:
            log.error('No sig in version file')

    def _load_json(self):
        """Downloads version file and converts to dict"""
        log.debug('Loading version file...')
        v = urllib2.urlopen(self.version_url)
        self.json_data = json.loads(v.read())

    def _get_highest_version(self, name):
        """Parses version file and returns the highest version number.

        Args:
            name (str): name of file to search for updates

        Returns:
            (str) Highest version number

        """
        highest_version = sorted(self.json_data['y_data'][name].keys())[-1]
        log.debug('Highest version: {}'.format(highest_version))
        return highest_version

    def _get_file_url(self, name, version):
        """Returns url for given name & version combo

        Args:
            name (str): name of file to get url for

            version (str): version of file to get url for

        Returns:
            (str) Url
        """
        url = self.json_data['y_data'][name][version]['url']
        log.debug('Update url: {}'.format(url))
        return url

    def _get_filename_with_extension(self, name, version):
        """Gets full filename for given name & version combo

        Args:
            name (str): name of file to get full filename for

            version (str): version of file to get full filename for

        Returns:
            (str) Filename with extension
        """
        # Parses version file and returns filename
        filename = self.json_data['y_data'][name][version]['filename']
        log.debug("Filename: {}".format(filename))
        return filename

    def _fix_update_url(self, url):
        """Adds trailing slash to urls provided in config if
        not already present

        Args:
            url (str): url to process

        Returns:
            (str) Url with trailing slash
        """
        # Adds trailing slash to end of url
        # if not already provided
        if not url.endswith('/'):
            return url + '/'
        return url
